from abc import ABC, abstractmethod


class HtmlGeneratorConfiguration(ABC):
    """HTML generator configuration interface."""

    @abstractmethod
    def get_markdown_tex_support(self) -> bool:
        """Get whether markdown TeX rendering is supported.

        :return: Markdown TeX supported
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def get_line_length(self) -> int:
        """Get line length.

        :return: Size in pixels
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def get_markdown_use_monospace_font(self) -> bool:
        """Get whether to use a monospace font for the markdown render.

        :return: Using monospace font
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def get_use_monospace_font(self) -> bool:
        """Get whether to use a monospace font.

        :return: Using monospace font
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def get_font_size(self) -> int:
        """Get font size.

        :return: Size
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def get_high_contrast(self) -> bool:
        """Get whether high contrast output should be used for screen rendering.

        :return: Whether visible
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def get_markdown_render_monospace_font_ratio(self) -> float:
        """Get the adjustment in size from proportional to fixed width font.

        :return: Ratio
        :rtype: float
        """
        raise NotImplementedError()

    @abstractmethod
    def get_editor_header_bar_visible_for_window_state(self) -> bool:
        """Get whether the header bar is configured visible for the window maximised state."

        :return: Whether visible
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def get_toolbar_underlay_padding_height(self) -> int:
        """Fetch the height of padding to account for overlay header bars.

        :return: Height in pixels
        :rtype: int
        """
        raise NotImplementedError()



from gi.repository import Adw, Gio, GLib

from enum import StrEnum
import logging
from typing import Optional

import iotas.const as const

settings = Gio.Settings.new(const.APP_ID)

STYLE = "style-variant"
WINDOW_SIZE = "window-size"
FIRST_START = "first-start"
FONT_SIZE = "font-size"
LINE_LENGTH = "line-length"
USE_MONOSPACE_FONT = "use-monospace-font"
EDITOR_THEME = "editor-theme"
EDITOR_FORMATTING_BAR_VISIBILTY = "editor-formatting-bar-visibility"
EDITOR_HEADER_BAR_VISIBILTY = "editor-header-bar-visibility"
BACKUP_NOTE_EXTENSION = "backup-note-extension"
PERSIST_SIDEBAR = "persist-sidebar"
INDEX_CATEGORY_STYLE = "index-category-style"

NEXTCLOUD_ENDPOINT = "nextcloud-endpoint"
NEXTCLOUD_USERNAME = "nextcloud-username"
NEXTCLOUD_PRUNE_THRESHOLD = "nextcloud-prune-threshold"
SYNC_INTERVAL = "sync-interval"
NETWORK_TIMEOUT = "network-timeout"

SPELLING_ENABLED = "spelling-enabled"
SPELLING_LANGUAGE = "spelling-language"

SHOW_STARTUP_SECRET_SERVICE_FAILURE = "show-startup-secret-service-failure"
SHOW_SYNCING_DEBUG_NOTIFICATION = "show-syncing-debug-notification"

MARKDOWN_RENDER = "markdown-render-enabled"
MARKDOWN_DETECT_SYNTAX = "markdown-syntax-detection-enabled"
MARKDOWN_KEEP_WEBKIT_PROCESS = "markdown-keep-webkit-process"
MARKDOWN_TEX_SUPPORT = "markdown-tex-support"
MARKDOWN_USE_MONOSPACE_FONT = "markdown-use-monospace-font"
MARKDOWN_RENDER_MONOSPACE_FONT_RATIO = "markdown-render-proportional-to-monospace-font-ratio"
MARKDOWN_DEFAULT_TO_RENDER = "markdown-default-to-render"

LAST_LAUNCHED_VERSION = "last-launched-version"
LAST_EXPORT_DIRECTORY = "last-export-directory"
EXTRA_EXPORT_FORMATS = "extra-export-formats"
SHOW_EXTENDED_PREFERENCES = "show-extended-preferences"


class HeaderBarVisibility(StrEnum):
    """Top or bottom bar visibility."""

    ALWAYS_VISIBLE = "always-visible"
    AUTO_HIDE = "auto-hide"
    AUTO_HIDE_FULLSCREEN_ONLY = "auto-hide-fullscreen-only"
    DISABLED = "disabled"


def get_window_size() -> GLib.Variant:
    """Get window size.

    :return: The size
    :rtype: GLib.Variant
    """
    return settings.get_value(WINDOW_SIZE)


def set_window_size(width: int, height: int) -> None:
    """Set window size.

    :param int width: Width
    :param int height: Height
    """
    g_variant = GLib.Variant("ai", (width, height))
    settings.set_value(WINDOW_SIZE, g_variant)


def get_first_start() -> bool:
    """Get whether doing first startup.

    :return: First startup
    :rtype: bool
    """
    return settings.get_value(FIRST_START)


def set_first_start(value: bool) -> None:
    """Set whether first start.

    :param bool value: New value
    """
    settings.set_boolean(FIRST_START, value)


def get_use_monospace_font() -> bool:
    """Get whether to use a monospace font.

    :return: Using monospace font
    :rtype: bool
    """
    return settings.get_value(USE_MONOSPACE_FONT)


def set_use_monospace_font(value: bool) -> None:
    """Set whether to use a monospace font.

    :param bool value: New value
    """
    settings.set_boolean(USE_MONOSPACE_FONT, value)


def get_font_size() -> int:
    """Get font size.

    :return: Size
    :rtype: int
    """
    return settings.get_int(FONT_SIZE)


def set_font_size(value: int) -> None:
    """Set font size.

    :param int value: New value
    """
    settings.set_int(FONT_SIZE, value)


def get_default_font_size() -> int:
    """Get default font size.

    :return: Default size
    :rtype: int
    """
    return settings.get_default_value(FONT_SIZE).get_int32()


def get_line_length() -> int:
    """Get line length.

    :return: Size in pixels
    :rtype: int
    """
    return settings.get_int(LINE_LENGTH)


def set_line_length(value: int) -> None:
    """Set line length.

    :param int value: New value
    """
    settings.set_int(LINE_LENGTH, value)


def get_default_line_length() -> int:
    """Get default line length.

    :return: Default length
    :rtype: int
    """
    return settings.get_default_value(LINE_LENGTH).get_int32()


def get_line_length_max() -> int:
    """Get line length maximum.

    :return: Size in pixels
    :rtype: int
    """
    return settings.get_property("settings-schema").get_key(LINE_LENGTH).get_range()[1][-1]


def get_spelling_enabled() -> bool:
    """Get whether spelling enabled.

    :return: Spelling enabled
    :rtype: bool
    """
    return settings.get_value(SPELLING_ENABLED)


def set_spelling_enabled(value: bool) -> None:
    """Set spelling enabled.

    :param bool value: New value
    """
    settings.set_boolean(SPELLING_ENABLED, value)


def get_spelling_language() -> Optional[str]:
    """Get spelling language.

    :return: Language tag or None if empty
    :rtype: str
    """
    lang = settings.get_string(SPELLING_LANGUAGE)
    if lang.strip() == "":
        lang = None
    return lang


def set_spelling_language(value: str) -> None:
    """Set spelling language.

    :param str value: New value
    """
    settings.set_string(SPELLING_LANGUAGE, value)


def get_style() -> str:
    """Get style.

    :return: Style
    :rtype: str
    """
    return settings.get_string(STYLE)


def set_style(value: str) -> None:
    """Set style.

    :param str value: New value
    """
    settings.set_string(STYLE, value)


def get_sync_interval() -> int:
    """Get sync interval.

    :return: Interval
    :rtype: int
    """
    return settings.get_int(SYNC_INTERVAL)


def set_sync_interval(value: int) -> None:
    """Set sync interval.

    :param int value: New value
    """
    settings.set_int(SYNC_INTERVAL, value)


def get_index_category_style() -> str:
    """Get the index category label style.

    :return: The style
    :rtype: str
    """
    return settings.get_string(INDEX_CATEGORY_STYLE)


def set_index_category_style(value: str) -> None:
    """Set the index category style.

    :param str value: New value
    """
    settings.set_string(INDEX_CATEGORY_STYLE, value)


def get_pin_sidebar() -> bool:
    """Get whether to pin the index sidebar (in large windows).

    :return: Whether to hide
    :rtype: bool
    """
    return settings.get_value(PERSIST_SIDEBAR)


def set_pin_sidebar(value: bool) -> None:
    """Set whether to pin the index sidebar (in large windows).

    :param bool value: New value
    """
    settings.set_boolean(PERSIST_SIDEBAR, value)


def get_editor_theme() -> str:
    """Get editor theme.

    :return: Theme
    :rtype: str
    """
    return settings.get_string(EDITOR_THEME)


def set_editor_theme(value: str) -> None:
    """Set editor theme.

    :param str value: New value
    """
    settings.set_string(EDITOR_THEME, value)


def get_editor_formatting_bar_visibility() -> Optional[HeaderBarVisibility]:
    """Get the formatting bar visibility.

    :return: Visibility, or None
    :rtype: HeaderBarVisibility, optional
    """
    setting_value = settings.get_string(EDITOR_FORMATTING_BAR_VISIBILTY)
    try:
        obj = HeaderBarVisibility(setting_value)
    except ValueError as e:
        logging.warning(f"Couldn't parse formatting bar visibility from '{setting_value}': {e}")
        return None
    else:
        return obj


def get_editor_formatting_bar_visible_for_window_state() -> bool:
    """Get whether the formatting bar is configured visible for the window maximised state."

    :return: Whether visible
    :rtype: bool
    """
    setting_value = settings.get_string(EDITOR_FORMATTING_BAR_VISIBILTY)
    try:
        obj = HeaderBarVisibility(setting_value)
    except ValueError as e:
        logging.warning(f"Couldn't parse formatting bar visibility from '{setting_value}': {e}")
        return False
    else:
        visible = False
        if get_markdown_detect_syntax():
            if obj == HeaderBarVisibility.ALWAYS_VISIBLE:
                visible = True
            elif obj == HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY:
                app = Gio.Application.get_default()
                if not app:
                    logging.error("Couldn't get app?")
                    return False
                window = app.get_active_window()
                visible = not window.is_fullscreen()
        return visible


def set_editor_formatting_bar_visibility(value: HeaderBarVisibility) -> None:
    """Set the formatting bar visibility.

    :param HeaderBarVisibility value: New value
    """
    settings.set_string(EDITOR_FORMATTING_BAR_VISIBILTY, str(value))


def get_editor_header_bar_visibility() -> Optional[HeaderBarVisibility]:
    """Get the header bar visibility.

    :return: The visibility setting
    :rtype: HeaderBarVisibility
    """
    setting_value = settings.get_string(EDITOR_HEADER_BAR_VISIBILTY)
    try:
        obj = HeaderBarVisibility(setting_value)
    except ValueError as e:
        logging.warning(f"Couldn't parse header bar visibility from '{setting_value}': {e}")
        return None
    else:
        return obj


def get_editor_header_bar_visible_for_window_state() -> bool:
    """Get whether the header bar is configured visible for the window maximised state."

    :return: Whether visible
    :rtype: bool
    """
    setting_value = settings.get_string(EDITOR_HEADER_BAR_VISIBILTY)
    try:
        obj = HeaderBarVisibility(setting_value)
    except ValueError as e:
        logging.warning(f"Couldn't parse header bar visibility from '{setting_value}': {e}")
        return False
    else:
        visible = False
        if obj == HeaderBarVisibility.ALWAYS_VISIBLE:
            visible = True
        elif obj == HeaderBarVisibility.AUTO_HIDE_FULLSCREEN_ONLY:
            app = Gio.Application.get_default()
            if not app:
                logging.error("Couldn't get app?")
                return False
            window = app.get_active_window()
            visible = not window.is_fullscreen()
        return visible


def set_editor_header_bar_visibility(value: HeaderBarVisibility) -> None:
    """Set the header bar visibility.

    :param HeaderBarVisibility value: New value
    """
    settings.set_string(EDITOR_HEADER_BAR_VISIBILTY, str(value))


def get_markdown_render_enabled() -> bool:
    """Get markdown render is enabled.

    :return: Markdown render enabled
    :rtype: bool
    """
    return settings.get_value(MARKDOWN_RENDER)


def set_markdown_render_enabled(value: bool) -> None:
    """Set markdown render enabled.

    :param bool value: New value
    """
    settings.set_boolean(MARKDOWN_RENDER, value)


def get_markdown_detect_syntax() -> bool:
    """Get markdown syntax detection is enabled.

    :return: Enabled
    :rtype: bool
    """
    return settings.get_value(MARKDOWN_DETECT_SYNTAX)


def set_markdown_detect_syntax(value: bool) -> None:
    """Set markdown syntax detection is enabled.

    :param bool value: New value
    """
    settings.set_boolean(MARKDOWN_DETECT_SYNTAX, value)


def get_markdown_keep_webkit_process() -> bool:
    """Get markdown WebKit process being retained.

    :return: WebKit process being retained
    :rtype: bool
    """
    return settings.get_value(MARKDOWN_KEEP_WEBKIT_PROCESS)


def set_markdown_keep_webkit_process(value: bool) -> None:
    """Set markdown WebKit process being retained.

    :param bool value: New value
    """
    settings.set_boolean(MARKDOWN_KEEP_WEBKIT_PROCESS, value)


def get_markdown_tex_support() -> bool:
    """Get whether markdown TeX rendering is supported.

    :return: Markdown TeX supported
    :rtype: bool
    """
    return settings.get_value(MARKDOWN_TEX_SUPPORT)


def set_markdown_tex_support(value: bool) -> None:
    """Set whether markdown TeX rendering is supported.

    :param bool value: New value
    """
    settings.set_boolean(MARKDOWN_TEX_SUPPORT, value)


def get_markdown_use_monospace_font() -> bool:
    """Get whether to use a monospace font for the markdown render.

    :return: Using monospace font
    :rtype: bool
    """
    return settings.get_value(MARKDOWN_USE_MONOSPACE_FONT)


def set_markdown_use_monospace_font(value: bool) -> None:
    """Set whether to use a monospace font for the markdown render.

    :param bool value: New value
    """
    settings.set_boolean(MARKDOWN_USE_MONOSPACE_FONT, value)


def set_markdown_render_monospace_font_ratio(value: float) -> None:
    """Set the adjustment in size from proportional to fixed width font.

    :param bool value: New value
    """
    settings.set_double(MARKDOWN_RENDER_MONOSPACE_FONT_RATIO, value)


def get_markdown_render_monospace_font_ratio() -> float:
    """Get the adjustment in size from proportional to fixed width font.

    :return: Ratio
    :rtype: float
    """
    return settings.get_double(MARKDOWN_RENDER_MONOSPACE_FONT_RATIO)


def get_markdown_default_to_render() -> bool:
    """Get whether to render the markdown when opening the note.

    :return: Defaulting to render
    :rtype: bool
    """
    return settings.get_value(MARKDOWN_DEFAULT_TO_RENDER)


def set_markdown_default_to_render(value: bool) -> None:
    """Set whether to render the markdown when opening the note.

    :param bool value: New value
    """
    settings.set_boolean(MARKDOWN_DEFAULT_TO_RENDER, value)


def get_nextcloud_endpoint() -> str:
    """Get Nextcloud endpoint.

    :return: Endpoint
    :rtype: str
    """
    return settings.get_string(NEXTCLOUD_ENDPOINT)


def set_nextcloud_endpoint(value: str) -> None:
    """Set Nextcloud endpoint.

    :param str value: New value
    """
    settings.set_string(NEXTCLOUD_ENDPOINT, value)


def get_nextcloud_username() -> str:
    """Get Nextcloud username.

    :return: Username
    :rtype: str
    """
    return settings.get_string(NEXTCLOUD_USERNAME)


def set_nextcloud_username(value: str) -> None:
    """Set Nextcloud username.

    :param str value: New value
    """
    settings.set_string(NEXTCLOUD_USERNAME, value)


def nextcloud_sync_configured() -> bool:
    """Get whether sync with Nextcloud is configured.

    This does not mean authentication has been successful this session.

    :return: Whether configured
    :rtype: bool
    """
    sync_username = get_nextcloud_username()
    sync_endpoint = get_nextcloud_endpoint()
    return sync_username != "" and sync_endpoint != ""


def get_nextcloud_prune_threshold() -> int:
    """Get Nextcloud prune threshold.

    :return: Threshold
    :rtype: int
    """
    return settings.get_int(NEXTCLOUD_PRUNE_THRESHOLD)


def set_nextcloud_prune_threshold(value: int) -> None:
    """Set Nextcloud prune threshold.

    :param int value: New value
    """
    settings.set_int(NEXTCLOUD_PRUNE_THRESHOLD, value)


def get_show_startup_secret_service_failure() -> bool:
    """Get to show Secret Service failure at startup.

    :return: Whether to show
    :rtype: bool
    """
    return settings.get_value(SHOW_STARTUP_SECRET_SERVICE_FAILURE)


def get_show_syncing_debug_notification() -> bool:
    """Get whether to show sync debug notifications.

    :return: Whether to show
    :rtype: bool
    """
    return settings.get_value(SHOW_SYNCING_DEBUG_NOTIFICATION)


def get_backup_note_extension() -> str:
    """Get file extension for backed up notes.

    :return: Extension
    :rtype: str
    """
    return settings.get_string(BACKUP_NOTE_EXTENSION)


def set_backup_note_extension(value: str) -> None:
    """Set file extension for backed up notes.

    :param str value: New value
    """
    settings.set_string(BACKUP_NOTE_EXTENSION, value)


def get_last_launched_version() -> str:
    """Get the last version which was run.

    :return: Version
    :rtype: str
    """
    return settings.get_string(LAST_LAUNCHED_VERSION)


def set_last_launched_version(value: str) -> None:
    """Set the last version which was run.

    :param str value: New value
    """
    settings.set_string(LAST_LAUNCHED_VERSION, value)


def get_last_export_directory() -> str:
    """Get the last export directory.

    :return: Directory
    :rtype: str
    """
    return settings.get_string(LAST_EXPORT_DIRECTORY)


def set_last_export_directory(value: str) -> None:
    """Set the last export directory.

    :param str value: New value
    """
    settings.set_string(LAST_EXPORT_DIRECTORY, value)


def get_extra_export_formats() -> str:
    """Get any extra export formats.

    :return: Raw formats string
    :rtype: str
    """
    return settings.get_string(EXTRA_EXPORT_FORMATS)


def set_extra_export_formats(value: str) -> None:
    """Set extra export formats.

    :param str value: New value
    """
    settings.set_string(EXTRA_EXPORT_FORMATS, value)


def get_show_extended_preferences() -> bool:
    """Get whether to show a great many preferences.

    :return: Whether to show
    :rtype: bool
    """
    return settings.get_value(SHOW_EXTENDED_PREFERENCES)


def set_show_extended_preferences(value: bool) -> None:
    """Set whether to show a great many preferences.

    :param bool value: New value
    """
    settings.set_boolean(SHOW_EXTENDED_PREFERENCES, value)


def get_network_timeout() -> float:
    """Get the network timeout.

    :return: Timeout in seconds
    :rtype: float
    """
    return settings.get_double(NETWORK_TIMEOUT)


def set_network_timeout(value: float) -> None:
    """Set the network timeout.

    :param int value: New value
    """
    settings.set_double(NETWORK_TIMEOUT, value)


# Piped through the configuration manager for now as a result of the HTML generator using it
# directly as configuration.
def get_toolbar_underlay_padding_height() -> int:
    """Fetch the height of padding to account for overlay header bars.

    :return: Height in pixels
    :rtype: int
    """
    return const.HEADER_BAR_HEIGHT


# Piped through the configuration manager for now as a result of the HTML generator using it
# directly as configuration.
def get_high_contrast() -> bool:
    """Get whether high contrast output should be used for screen rendering.

    :return: Whether visible
    :rtype: bool
    """
    return Adw.StyleManager.get_default().get_high_contrast()


# TODO remove ~early 2025 #####################################################

# Retaining to migrate
MARKDOWN_SYNTAX_HIGHLIGHTING = "markdown-syntax-highlighting-enabled"
HIDE_EDITOR_HEADERBAR = "auto-hide-editor-headerbar"
HIDE_HEADERBAR_WHEN_FULLSCREEN = "hide-editor-headerbar-when-fullscreen"


def get_markdown_syntax_hightlighting_enabled() -> bool:
    """Get markdown syntax highlighting is enabled.

    :return: Highlighting enabled
    :rtype: bool
    """
    return settings.get_value(MARKDOWN_SYNTAX_HIGHLIGHTING)


def get_hide_editor_headerbar() -> bool:
    """Get whether to hide the editor headerbar.

    :return: Whether to hide
    :rtype: bool
    """
    return settings.get_value(HIDE_EDITOR_HEADERBAR)


def get_hide_editor_headerbar_when_fullscreen() -> bool:
    """Get whether to hide the editor headerbar when fullscreen.

    :return: Whether to hide
    :rtype: bool
    """
    return settings.get_value(HIDE_HEADERBAR_WHEN_FULLSCREEN)


###############################################################################



from markdown_it import MarkdownIt
from markdown_it.renderer import RendererProtocol
from markdown_it.token import Token
from markdown_it.utils import EnvType, OptionsDict
from typing import Sequence


def img_lazyload_plugin(md: MarkdownIt) -> None:
    """Plugin ported from
    `mdit/plug-img-lazyload <https://github.com/mdit-plugins/mdit-plugins/blob/main/packages/img-lazyload`__.

    It adds the lazy loading attribute to images.
    """

    def render(
        self: RendererProtocol,
        tokens: Sequence[Token],
        idx: int,
        options: OptionsDict,
        env: EnvType,
    ) -> str:
        tokens[idx].attrSet("loading", "lazy")

        return self.renderToken(tokens, idx, options, env)

    md.add_render_rule("image", render)



from gettext import gettext as _
from gi.repository import Adw, Gdk, Gio, GLib, Gtk, GObject

import json
import logging
import os
import re
from typing import Any, Callable, Optional

import iotas.config_manager
import iotas.const as const
from iotas.note import Note
from iotas.webkit_pdf_exporter import WebKitPdfExporter


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/export_dialog.ui")
class ExportDialog(Adw.Dialog):
    __gtype_name__ = "ExportDialog"

    _main_stack: Gtk.Stack = Gtk.Template.Child()
    _headerbar_stack: Gtk.Stack = Gtk.Template.Child()
    _format_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _working_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _success_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _failure_headerbar: Adw.HeaderBar = Gtk.Template.Child()
    _format_listbox: Gtk.ListBox = Gtk.Template.Child()
    _cancel_button: Gtk.Button = Gtk.Template.Child()
    _show_button: Gtk.Button = Gtk.Template.Child()
    _working: Gtk.Box = Gtk.Template.Child()
    _success: Gtk.Box = Gtk.Template.Child()
    _success_label: Gtk.Label = Gtk.Template.Child()
    _failure: Gtk.Box = Gtk.Template.Child()
    _failure_label: Gtk.Label = Gtk.Template.Child()

    def __init__(
        self,
        note: Note,
        html_generator,
        webkit_init_fn: Callable[[], None],
    ) -> None:
        super().__init__()
        self.__note = note
        self.__html_generator = html_generator
        self.__webkit_init = webkit_init_fn
        # Any typing for lazy loading
        self.__exporter: Optional[Any] = None
        self.__show_path: Optional[str] = None
        self.__custom_formats: dict[str, dict[str, str]] = {}

        controller = Gtk.EventControllerKey.new()
        controller.connect("key-pressed", self._on_key_pressed)
        self.add_controller(controller)

        GLib.idle_add(self.__setup)

    def __setup(self) -> None:
        self._cancel_button.grab_focus()
        self.__load_extra_formats()

    @Gtk.Template.Callback()
    def _row_activated(self, _listbox: Gtk.ListBox, row: Adw.ButtonRow) -> None:
        format_id = row.get_title().lower()
        if format_id in self.__custom_formats:
            extension = self.__custom_formats[format_id]["extension"]
            format = self.__custom_formats[format_id]["format"]
        else:
            extension = format_id
            format = format_id
        self.__determine_output_location(format, extension)

    @Gtk.Template.Callback()
    def _on_show_clicked(self, _button: Gtk.Button) -> None:
        launcher = Gtk.FileLauncher.new(Gio.File.new_for_uri(f"file://{self.__show_path}"))
        launcher.open_containing_folder()
        self.close()

    @Gtk.Template.Callback()
    def _on_cancel(self, _button: Gtk.Button) -> None:
        self.close()

    def _on_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if self._main_stack.get_visible_child() != self._working:
            if keyval in (Gdk.KEY_Down, Gdk.KEY_KP_Down):
                if self.get_property("focus-widget") == self._cancel_button:
                    row = self._format_listbox.get_row_at_index(0)
                    row.grab_focus()
                    return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def __on_finished(self, _obj: GObject.Object, out_format: str, show_path: str) -> None:
        self._main_stack.set_visible_child(self._success)
        self._headerbar_stack.set_visible_child(self._success_headerbar)
        # Translators: Description, {} is a format eg. PDF
        self._success_label.set_label(_("Exported to {}").format(out_format.upper()))
        self.__show_path = show_path

        # For some reason if I call grab focus directly from the idle it blocks clean exit
        def focus_show_button():
            self._show_button.grab_focus()

        GLib.idle_add(focus_show_button)

    def __on_failed(self, _obj: GObject.Object, out_format: str, reason: str) -> None:
        self._main_stack.set_visible_child(self._failure)
        self._headerbar_stack.set_visible_child(self._failure_headerbar)
        # Translators: Description, {} is a format eg. PDF
        message = _("Failed to export to {}").format(out_format.upper())
        if reason != "":
            message = message + "\n\n" + reason
        self._failure_label.set_label(message)

    def __determine_output_location(self, out_format: str, file_extension: str) -> None:
        self.__init_exporter()
        assert self.__exporter
        filesystem_writable_path = self.__get_writable_non_container_dir()
        if filesystem_writable_path:
            dialog = Gtk.FileDialog.new()
            dialog.set_initial_folder(Gio.File.new_for_path(filesystem_writable_path))
            filename = self.__exporter.build_default_filename(
                self.__note, out_format, file_extension
            )
            dialog.set_initial_name(filename)
            # Translators: Button
            dialog.set_accept_label(_("Export"))
            window = self.get_parent().get_root()
            dialog.save(
                window,
                None,
                self.__on_save_dialog_finish,
                out_format,
                file_extension,
            )
        else:
            logging.info(
                "Can't access user documents directory, exporting to location inside container"
            )
            self.__export(out_format, file_extension, location=None)

    def __on_save_dialog_finish(
        self, dialog: Gtk.FileDialog, task: Gio.Task, out_format: str, file_extension: str
    ) -> None:
        try:
            file = dialog.save_finish(task)
        except GLib.GError as e:
            logging.warning("Couldn't export note: %s", e.message)
            self.close()
        else:
            path = file.get_path()

            if out_format == "html":
                dir_path = os.path.abspath(os.path.join(path, os.pardir))
            else:
                dir_path = os.path.dirname(path)
            iotas.config_manager.set_last_export_directory(dir_path)

            self.__export(out_format, file_extension, path)

    def __export(self, out_format: str, file_extension: str, location: Optional[str]) -> None:
        assert self.__exporter
        self._main_stack.set_visible_child(self._working)
        self._headerbar_stack.set_visible_child(self._working_headerbar)
        self.__exporter.export(
            self.__note,
            out_format,
            file_extension,
            iotas.config_manager.get_markdown_tex_support(),
            location,
        )

    def __get_writable_non_container_dir(self) -> Optional[str]:
        last_dir = iotas.config_manager.get_last_export_directory()
        if last_dir != "" and self.__dir_exists_writable(last_dir):
            return last_dir
        elif self.__can_write_documents_dir():
            return self.__get_documents_dir()
        else:
            return None

    def __get_documents_dir(self) -> Optional[str]:
        return GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DOCUMENTS)

    def __can_write_documents_dir(self) -> bool:
        documents_dir = self.__get_documents_dir()
        if documents_dir is None:
            return False
        else:
            return self.__dir_exists_writable(documents_dir)

    def __dir_exists_writable(self, path: str) -> bool:
        return os.path.exists(path) and os.access(path, os.W_OK)

    def __init_exporter(self) -> None:
        # More lazyloading to quicken startup
        from iotas.exporter import Exporter

        self.__pdf_exporter = WebKitPdfExporter(
            self.__webkit_init(), iotas.config_manager.get_markdown_keep_webkit_process()
        )

        self.__exporter = Exporter(self.__pdf_exporter, self.__html_generator, const.PKGDATADIR)
        self.__exporter.connect("finished", self.__on_finished)
        self.__exporter.connect("failed", self.__on_failed)

    def __load_extra_formats(self) -> bool:
        raw_formats = iotas.config_manager.get_extra_export_formats()
        if raw_formats.strip() == "":
            return False
        try:
            formats_json = json.loads(raw_formats)
        except json.JSONDecodeError as e:
            logging.warning("Failed to load extra formats: %s", e.msg)
            return False

        if type(formats_json) is not list:
            logging.warning("Failed to load extra formats: JSON not being a list")
            return False

        added = False
        for item in formats_json:
            if type(item) is not dict:
                logging.warning("Failed to load extra format: item not being an object")
                continue
            if "pandocOutFormat" not in item:
                logging.warning("Failed to load extra format: missing pandocOutFormat")
                continue
            if "fileExtension" not in item:
                logging.warning("Failed to load extra format: missing fileExtension")
                continue

            out_format = item["pandocOutFormat"]
            extension = item["fileExtension"]
            if type(out_format) is not str or out_format.strip() == "":
                logging.warning("Failed to load extra format: pandocOutFormat is not a string")
                continue
            if type(extension) is not str or extension.strip() == "":
                logging.warning("Failed to load extra format: due to fileExtension is not a string")
                continue

            if re.match(r"^[\w_]+$", out_format) is None:
                logging.warning(
                    "Failed to load extra format: pandocOutFormat contains invalid characters"
                )
                continue
            if re.match(r"^[\w]+$", extension) is None:
                logging.warning(
                    "Failed to load extra format: fileExtension contains invalid characters"
                )
                continue

            logging.debug(f"Adding extra pandoc format {out_format} with extension {extension}")
            button = Adw.ButtonRow()
            button.set_title(extension.upper())
            button.set_end_icon_name("go-next-symbolic")
            self._format_listbox.append(button)

            format_id = out_format.lower()
            self.__custom_formats[format_id] = {"format": out_format, "extension": extension}

            added = True

        return added



from gi.repository import GObject, GLib

import datetime
import logging
import os
import re
import shutil
from typing import Optional
import unicodedata

try:
    import pypandoc
except ModuleNotFoundError:
    pypandoc = None

from iotas.html_generator import HtmlGenerator
from iotas.note import Note
from iotas.pdf_exporter import PdfExporter


class Exporter(GObject.Object):

    __gsignals__ = {
        # Out format, out path
        "finished": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
        # Out format and reason
        "failed": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
    }

    def __init__(
        self, pdf_exporter: PdfExporter, html_generator: HtmlGenerator, app_data_path: str
    ) -> None:
        super().__init__()
        self.__pdf_exporter = pdf_exporter
        self.__pdf_exporter.set_callbacks(
            self.__on_pdf_export_finished, self.__on_pdf_export_failed
        )
        self.__html_generator = html_generator
        self.__app_data_path = app_data_path
        self.__active = False
        self.__in_error = False

    def export(
        self,
        note: Note,
        out_format: str,
        file_extension: str,
        supporting_tex: bool,
        user_location: Optional[str] = None,
    ) -> None:
        """Export note.

        :param Note note: Note to render
        :param str out_format: Export format
        :param str file_extension: File extension
        :param bool supporting_tex: TeX support
        :param str user_location: User chosen export location
        """
        self.__note = note
        self.__out_format = out_format
        self.__supporting_tex = supporting_tex
        if user_location:
            self.__location = user_location
        else:
            # Running with limited permissions in container, export to exports dir inside container
            # with automatic filename
            export_dir = os.path.join(GLib.get_user_data_dir(), "iotas", "exports")
            filename = self.build_default_filename(
                note, out_format, file_extension, add_timestamp=True
            )
            self.__location = os.path.join(export_dir, filename)
            if not os.path.exists(export_dir):
                try:
                    os.mkdir(export_dir)
                except OSError as e:
                    logging.warning(f"Failed to export {out_format} to {self.__location}: %s", e)
                    self.emit("failed", self.__out_format, e)
                    return

        if out_format == "pdf":
            self.__export_pdf()
        elif out_format == "md":
            self.__export_md()
        elif out_format == "html":
            self.__export_html()
        else:
            logging.info(f"Asking pandoc to export to {out_format}")
            self.__export_pandoc(out_format)

    def build_default_filename(
        self, note: Note, out_format: str, file_extension: str, add_timestamp: bool = False
    ) -> str:
        """Build an export filename for the note.

        :param Note note: Note to render
        :param str out_format: Export format
        :param str file_extension: File extension
        :param bool add_timestamp: Whether to prefix timestamp
        :return: Filename
        :rtype: str
        """
        filename = self.__sanitise_title_for_filename(note.title)
        if add_timestamp:
            ts = datetime.datetime.now()
            filename = ts.strftime("%Y-%m-%dT%H:%M:%S") + " " + filename
        if out_format != "html":
            filename += "." + file_extension
        return filename

    @GObject.Property(type=bool, default=False)
    def active(self) -> bool:
        return self.__active

    @active.setter
    def set_active(self, value: bool) -> None:
        self.__active = value

    def __on_pdf_export_finished(self) -> None:
        self.__active = False
        self.emit("finished", "pdf", self.__location)

    def __on_pdf_export_failed(self, error: str) -> None:
        self.__active = False
        self.emit("failed", "pdf", error)

    def __export_pandoc(self, out_format: str) -> None:
        self.__active = True
        if pypandoc:
            try:
                pypandoc.convert_text(
                    self.__note.content, out_format, format="gfm", outputfile=self.__location
                )
            except (RuntimeError, OSError) as e:
                logging.warning(f"Failed to export {out_format} to {self.__location}: %s", e)
                self.emit("failed", self.__out_format, str(e))
            else:
                logging.info(f"Exported {self.__out_format} to {self.__location}")
                self.emit("finished", self.__out_format, self.__location)
        else:
            logging.warning(
                f"Failed to export {out_format} to {self.__location}: "
                "Pypandoc is not installed on your system."
            )
            self.emit("failed", self.__out_format, "Pypandoc is not installed on your system.")
        self.__active = False

    def __export_pdf(self) -> None:
        self.__active = True
        self.__pdf_exporter.export(self.__note, self.__location)

    def __export_md(self) -> None:
        self.__active = True

        try:
            with open(self.__location, "w") as f:
                f.write(self.__note.content)
        except OSError as e:
            self.emit("failed", self.__out_format, e)
            logging.warning("Failed to export MD: %s", e)
            self.__active = False
            return

        self.__active = False
        logging.info(f"Exported {self.__out_format} to {self.__location}")
        self.emit("finished", self.__out_format, self.__location)

    def __export_html(self) -> None:
        self.__active = True

        (content, _) = self.__html_generator.generate(
            self.__note, searching=False, export_format="html"
        )

        if os.path.exists(self.__location):
            logging.debug(f"Removing existing {self.__location}")
            try:
                shutil.rmtree(self.__location)
            except OSError as e:
                self.emit("failed", self.__out_format, e)
                logging.warning("Failed to replace dir for HTML export: %s", e)
                self.__active = False
                return

        try:
            os.mkdir(self.__location)
        except OSError as e:
            self.emit("failed", self.__out_format, e)
            logging.warning("Failed to make dir for HTML export: %s", e)
            self.__active = False
            return

        index_filename = os.path.join(self.__location, "index.html")

        try:
            with open(index_filename, "w") as f:
                f.write(content)
        except OSError as e:
            self.emit("failed", self.__out_format, e)
            logging.warning("Failed to export HTML: %s", e)
            self.__active = False
            return

        css_dir = os.path.join(self.__location, "css")

        try:
            os.mkdir(css_dir)
        except OSError as e:
            self.emit("failed", self.__out_format, e)
            logging.warning("Failed to export HTML: %s", e)
            self.__active = False
            return

        dest_file = os.path.join(css_dir, os.path.basename(HtmlGenerator.RESOURCE_CSS_PATH))

        try:
            shutil.copyfile(f"{self.__app_data_path}/{HtmlGenerator.RESOURCE_CSS_PATH}", dest_file)
        except OSError as e:
            self.emit("failed", self.__out_format, e)
            logging.warning("Failed to export HTML: %s", e)
            self.__active = False
            return

        if self.__supporting_tex:
            dest_file = os.path.join(
                css_dir, os.path.basename(HtmlGenerator.RESOURCE_KATEX_CSS_PATH)
            )

            try:
                shutil.copyfile(
                    f"{self.__app_data_path}/{HtmlGenerator.RESOURCE_KATEX_CSS_PATH}", dest_file
                )
            except OSError as e:
                self.emit("failed", self.__out_format, e)
                logging.warning("Failed to export HTML: %s", e)
                self.__active = False
                return

            js_dir = os.path.join(self.__location, "js")

            try:
                os.mkdir(js_dir)
            except OSError as e:
                self.emit("failed", self.__out_format, e)
                logging.warning("Failed to export HTML: %s", e)
                self.__active = False
                return

            dest_file = os.path.join(js_dir, os.path.basename(HtmlGenerator.RESOURCE_KATEX_JS_PATH))

            try:
                shutil.copyfile(
                    f"{self.__app_data_path}/{HtmlGenerator.RESOURCE_KATEX_JS_PATH}", dest_file
                )
            except OSError as e:
                self.emit("failed", self.__out_format, e)
                logging.warning("Failed to export HTML: %s", e)
                self.__active = False
                return

        self.__active = False
        logging.info(f"Exported {self.__out_format} to {self.__location}")
        self.emit("finished", self.__out_format, self.__location)

    def __sanitise_title_for_filename(self, title: str) -> str:
        """For synced notes the server has already done the sanitising for us. This is for
        local-only instances.
        """
        value = unicodedata.normalize("NFKC", str(title))
        return re.sub(r"[^\w\s\.-]", "", value).strip()
